/*
 * Copyright (c) 2003 - 2007, Nils R. Weller
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * This file contains the functions and data needed to parse GNU-C-style
 * __attribute__ arguments
 */ 

#include "attribute.h"
#include <string.h>
#include <stdlib.h>
#include "token.h"
#include "misc.h"
#include "expr.h"
#include "error.h"
#include "debug.h"
#include "n_libc.h"

#define A_IMPL		1
#define A_UNIMPL	2
#define A_IGNORED	3

static struct attrent {
	char	*name;
	int	value;
	int	type;
	int	*args;
	int	nargs; /* 0 = no args, -1 = variable number */
} attrtab[] = {
	{ "alias", ATTRF_ALIAS, A_UNIMPL, 0, 0 },
	{ "aligned", ATTRS_ALIGNED, A_UNIMPL, 0, 0 },
	{ "always_inline", ATTRF_ALWAYS_INLINE, A_UNIMPL, 0, 0 },
	{ "cdecl", ATTRF_CDECL, A_UNIMPL, 0, 0 },
	{ "cleanup", ATTRV_CLEANUP, A_UNIMPL, 0, 0 },
	{ "common", ATTRV_COMMON, A_UNIMPL, 0, 0 },
	{ "const", ATTRF_CONST	, A_IGNORED, 0, 0 },	
	{ "constructor", ATTRF_CONSTRUCTOR, A_UNIMPL, 0, 0 },
	{ "deprecated", ATTRS_DEPRECATED, A_UNIMPL, 0, 0 },
	{ "destructor", ATTRF_DESTRUCTOR, A_UNIMPL, 0, 0 },
	{ "dllimport", ATTRV_DLLIMPORT, A_UNIMPL, 0, 0 },	
	{ "dllexport", ATTRV_DLLEXPORT, A_UNIMPL, 0, 0 },	
	{ "eightbit_data", ATTRF_EIGHTBIT_DATA, A_UNIMPL, 0, 0 },	
	{ "far", ATTRF_FAR, A_UNIMPL, 0, 0 },
	{ "format", ATTRF_FORMAT, A_UNIMPL, 0, 0 },
	{ "format_arg", ATTRF_FORMAT_ARG, A_UNIMPL, 0, 0 },
	{ "function_vector", ATTRF_FUNCTION_VECTOR, A_UNIMPL, 0, 0 },
	{ "interrupt", ATTRF_INTERRUPT, A_UNIMPL, 0, 0 },	
	{ "interrupt_handler", ATTRF_INTERRUPT_HANDLER, A_UNIMPL, 0, 0 },
	{ "longcall", ATTRF_LONGCALL, A_UNIMPL, 0, 0 },
	{ "malloc", ATTRF_MALLOC, A_UNIMPL, 0, 0 },
	{ "may_alias", ATTRS_MAY_ALIAS, A_UNIMPL, 0, 0 },	
	{ "mode", ATTRV_MODE, A_UNIMPL, 0, 0 },
	{ "model", ATTRV_MODEL, A_UNIMPL, 0, 0 },
	{ "naked", ATTRF_NAKED, A_UNIMPL, 0, 0 },
	{ "near", ATTRF_NEAR, A_UNIMPL, 0, 0 },
	{ "no_instrument_function",
		ATTRF_NO_INSTRUMENT_FUNCTION, A_UNIMPL, 0, 0 },
	{ "nocommon", ATTRV_NOCOMMON, A_UNIMPL, 0, 0 },
	{ "noinline", ATTRF_NOINLINE, A_UNIMPL, 0, 0 },
	{ "nonnull", ATTRF_NONNULL, A_IGNORED, 0, 0 },
	{ "nothrow", ATTRF_NOTHROW, A_IGNORED, 0, 0 },
	{ "packed", ATTRS_PACKED, A_UNIMPL, 0, 0 },
	{ "pure", ATTRF_PURE, A_UNIMPL, 0, 0 },
	{ "regparm", ATTRF_REGPARM, A_UNIMPL, 0, 0 },
	{ "section", ATTRF_SECTION, A_UNIMPL, 0, 0 },
	{ "shared", ATTRV_SHARED, A_UNIMPL, 0, 0 },
	{ "shortcall", ATTRF_SHORTCALL, A_UNIMPL, 0, 0 },
	{ "signal", ATTRF_SIGNAL, A_UNIMPL, 0, 0 },
	{ "sp_switch", ATTRF_SP_SWITCH, A_UNIMPL, 0, 0 },
	{ "stdcall", ATTRF_STDCALL, A_UNIMPL, 0, 0 },
	{ "tiny_data", ATTRF_TINY_DATA, A_UNIMPL, 0, 0 },
	{ "tls_model", ATTRV_TLS_MODEL, A_UNIMPL, 0, 0 },
	{ "transparent_union", ATTRS_TRANSPARENT_UNION, A_UNIMPL, 0, 0 },
	{ "trap_exit", ATTRF_TRAP_EXIT, A_UNIMPL, 0, 0 },
 	{ "unused", ATTRF_UNUSED, A_IGNORED, 0, 0 },
	{ "used", ATTRF_USED, A_IGNORED, 0, 0 },
	{ "vector_size", ATTRV_VECTOR_SIZE, A_UNIMPL, 0, 0 },
	{ "visibility", ATTRF_VISIBILITY, A_UNIMPL, 0, 0 },
	{ "weak", ATTRF_WEAK, A_UNIMPL, 0, 0 },
	{ 0, 0, 0, 0, 0 }
};

static struct attrent *
find_attrent(char *name) {
	int	i;
	size_t	len = 0;
	char	*p = NULL;

	/*
	 * All attribute specifiers might either have the form ``attr'' or
	 * ``__attr__'', so we have to determine which is the case here
	 */
	if (*name == '_') {
		if (name[1] != '_') {
			return NULL;
		}
		p = name + 2;
		if ((len = strlen(p)) < 3) {
			/* Bogus attribute, becase it can't match __*X*__ */
			return NULL;
		}
		if (p[len - 1] != '_' || p[len - 2] != '_') {
			return NULL;
		}
	}

	for (i = 0; attrtab[i].name != NULL; ++i) {
		if (p == NULL) {
			/* No trailing __ to take into account */
			if (strcmp(attrtab[i].name, name) == 0) {
				/* Found */
				return &attrtab[i];
			}
		} else if (strncmp(attrtab[i].name, p, len - 2) == 0) {
			/* Found */
			return &attrtab[i];
		}
	}
	/* Not found */
	return NULL;
}
	

static void
free_attrib(struct attrib *attr) {
	struct attrib	*at;

	while (attr) {
		at = attr;
		attr = attr->next;
		free(at);
	}
}


static struct attrib *
alloc_attrib(void) {
	struct attrib	*ret = n_xmalloc(sizeof *ret);
	static struct attrib	nullattrib;
	*ret = nullattrib;
	return ret;
}	

struct attrib *
get_attribute(struct token **tok) {
	struct token	*t = *tok;
	struct attrib	*ret = NULL;
	struct attrent	*at;
	int		err;

#if 0
	/* XXX temporary */
	*tok = ignore_attr(*tok);
	return NULL;
#endif

	/* 
	 * Function might be called at ``__attribute__'' so as to avoid
	 * cluttering the caller with next_token() calls
	 */
	if (t->type == TOK_KEY_ATTRIBUTE) {
		if (next_token(&t) != 0) {
			return NULL;
		}
	}

	/* Every attribute must look like ``__attribute__((attr-list))'' */
	if (t->type != TOK_PAREN_OPEN) {
		errorfl(t, "Syntax error at %s", t->ascii);
		*tok = t;
		return NULL;
	}

	if (next_token(&t) != 0) {
		*tok = t;
		return NULL;
	}

	if (t->type != TOK_PAREN_OPEN) {
		errorfl(t, "Syntax error at %s", t->ascii);
		*tok = t;
		return NULL;
	}

	ret = n_xmalloc(sizeof *ret);

	err = 0;
	for (; t != NULL; t = t->next) {
		if (t->type == TOK_PAREN_CLOSE) {
			if (next_token(&t) != 0) {
				*tok = t;
				free_attrib(ret);
				return NULL;
			}

			/* ... )) */
			if (t->type != TOK_PAREN_CLOSE) {
				errorfl(t, "Syntax error at %s", t->ascii);
				*tok = t;
				free_attrib(ret);
				return NULL;
			}

			if (err) {
				return NULL;
			} else {
				*tok = t->next;
				return ret;
			}
		} else if (t->type == TOK_IDENTIFIER) {
			if ((at = find_attrent(t->data)) == NULL) {
#if 0				
				errorfl(t, "Unknown attribute");
#endif				
				err = 1;
			} else if (at->type == A_UNIMPL) {
#if 0
				errorfl(t, "Attribute not "
					"implemented - use gcc to compile "
					"this code!");
				err = 1;
#endif
			} else if (at->type != A_IGNORED) {
				/*
				 * Neither unimpl. nor ignored =
				 * implemented!
				 */ 
				/* XXX read arguments... */
			}
			if (at && at->value == ATTRV_MODE) {
				if (expect_token(&t, TOK_PAREN_OPEN, 1) != 0) {
					return NULL;
				}
				if (t->type != TOK_IDENTIFIER) {
					return NULL;
				}
				if (strcmp(t->ascii, "__DI__") == 0) {
					if (expect_token(&t, TOK_PAREN_CLOSE,
						1) != 0 ||
					expect_token(&t, TOK_PAREN_CLOSE, 1)
						!= 0) {
						return NULL;
					}
					*tok = t;
					ret = alloc_attrib();
					ret->code = ATTRV_MODE;
					ret->iarg = AMODE_DI; 
					return ret;
				} else {	
					err = 1;
				}	
			} else {
				err = 1;
			}	
			if (err) {
				*tok = ignore_attr(*tok);
				return NULL;
			}
		}
	}
	errorfl(*tok, "Unexpected end of file");
	*tok = t;
	return NULL;
}


struct token *
ignore_attr(struct token *tok) {
	int	parens = 0;

	if (tok == NULL) {
		return NULL;
	}
	if (tok->type != TOK_KEY_ATTRIBUTE) {
		return tok;
	}
	for (tok = tok->next; tok != NULL; tok = tok->next) {
		if (tok->type == TOK_PAREN_OPEN) {
			++parens;
		} else if (tok->type == TOK_PAREN_CLOSE) {
			if (--parens == 0) {
				return tok->next;
			}
		}
	}
	return NULL;
}

