/*
 * Copyright (c) 2003 - 2006, Nils R. Weller
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */
#ifndef TYPE_H
#define TYPE_H


struct decl;
struct type;
struct vreg;
struct token;
struct stack_block;

#include <stddef.h>

/* For structure definitions */
struct ty_struct {
	char			*tag;
	struct scope		*scope;
	struct scope		*parentscope;

	/* number of members */
	int			nmemb;

	/* e.g. __attribute__((packed)) */
	int			attrib;
	int			unnamed;
	int			incomplete;
	int			printed;
	int			is_union;
	int			alignment;
	unsigned int		references;
	size_t			size;
	struct decl		*flexible;
	struct ty_struct	*next;
};


/*
 * An array of ty_bit in ``struct type'' is used to describe a bitfield
 */
struct ty_bit {
	char			*name;
	int				numbits;
};

/*
 * Data structure to store string literal information. This is not
 * used by ``struct type'', but only by the ``struct token'' data
 * member, initializers, etc
 */
struct ty_string {
	char			*str; /* The string literal */
	unsigned long	count; /* .str%lu the `%lu' in asm output */
	size_t			size; /* Size of array (including null) */
	struct type		*ty;
	struct ty_string	*next;
};

struct num;

struct ty_float {
	struct num	*num;
	unsigned long	count;
	struct ty_float	*next;
};	

/* For enum definitions */
struct ty_enum {
	char *tag;
	struct {
		char		*name;
		struct decl	*dec;
	}    *members;
	int	 nmemb;
};

/* For function declarations */
#define FDTYPE_KR	1
#define FDTYPE_ISO	2

struct ty_func { 
	char		*name;
	struct scope	*scope;
	struct decl	*lastarg;
	int		nargs;	/* number of arguments */
	struct type	*ret;	/* return type */
	char		*asmname; /* assembler name */
	int		type;	/* K&R or ISO declaration */
	int		variadic; /* Variadic function? */
};


#define TN_ARRAY_OF		(1)
#define TN_POINTER_TO	(1 << 1)
#define TN_FUNCTION		(1 << 2)


/*
 * This structure describes the properties of a declarator, i.e.
 * ``pointer-to'', ``array-of-N'', ``function pointer'', etc
 */
struct type_node {
	/* Designates array of/pointer to/function */
	int					type;

	/* If this is an array, arrayarg specifies the number of elements */
	struct expr			*arrarg;
	/* 
	 * Number of elements for string constants, and after parsing
	 * arrays with constant size
	 */
	size_t				arrarg_const;

	/*
	 * If this is a pointer, ptrarg specifies whether the pointer is
	 * qualified (TOK_KEY_VOLATILE/CONST/RESTRICT) or not (0)
	 */
	int					ptrarg;

	/*
	 * If this is a function, tfunc specifies its arguments.
	 * This is a null pointer in case of ``T f(void)''
	 */
	struct ty_func		*tfunc;

	/* Next node in list */
	struct type_node	*next;

	/* Previous node in list */
	struct type_node	*prev;
};


struct type {
	int		line; /* line in source file */
	char		*file; /* name of source file */
	char		*name;
	int		is_func; /* is function declaration? */
	int		is_def;	/* is struct/func definition? */

	/* Structure/union/enum-specific data */
	int			incomplete; /* is incomplete */
	struct ty_struct	*tstruc;
	struct ty_enum		*tenum;

	/*
	 * tbit is used for bitfield members. A bitfield is designated by
	 * a struct member of type (u)int/bool with tbit != NULL. That is,
	 * struct ty_struc *t = lookup_struct("tag", SCOPE_NESTED);
	 * if (t != NULL) {
	 *     struct type *ty = t->members[0].dtype;
	 *     if (ty->code == TY_INT ||
	 *       ty->code == TY_UINT || 
	 *         ty->code == TY_BOOL) {
	 *         if (ty->tbit != NULL) {
	 *            fist member is bitfield!
	 *         }
	 *     }
	 *  }
	 * The same applies to unions
	 */
	struct ty_bit		*tbit;

	/* General data */
	int			code; /* integer, double, etc */
	int			implicit;/* implicit declaration? */
	int			sign; /* signed or unsigned */
	int			storage; /* storage duration */
	int			is_const; /* const? */
	int			is_volatile; /* volatile? */
	int			is_restrict; /* restrict? */
	int			is_inline; /* inline? */

	/*
	 * Declarator list. This is not used in 
	 * structure/... definitions!
	 */
	struct type_node	*tlist;			

	struct type_node	*tlist_tail;

};


#define CMPTY_SIGN	1
#define CMPTY_CONST	(1 << 2)
#define CMPTY_ALL	(CMPTY_SIGN|CMPTY_CONST)
#define CMPTY_ARRAYPTR	(1 << 3)
#define CMPTY_TENTDEC	(1 << 4)

struct ty_struct	*alloc_ty_struct(void);
struct ty_string	*alloc_ty_string(void);
struct ty_string	*make_ty_string(const char *, size_t);
unsigned long		ty_string_count(void);
struct ty_bit		*alloc_ty_bit(void);
struct ty_enum		*alloc_ty_enum(void);
struct ty_func		*alloc_ty_func(void);
struct type_node	*alloc_type_node(void);
struct type		*alloc_type(void);
int	compare_types(struct type *dest, struct type *src, int flag);
int		check_init_type(struct type *dest, struct expr *init);
void	copy_type(struct type *dest, const struct type *src, int full);
void	copy_tlist(struct type_node **dest, const struct type_node *tlist);
struct type		*make_basic_type(int code);
struct type		*make_array_type(int size);
void			append_typelist(struct type *t,
				int type, void *type_arg, struct ty_func *tf,
				struct token *tok);

char			*type_to_text(struct type *t);
int			is_arithmetic_type(struct type *t);
int			is_scalar_type(struct type *t);
int			is_floating_type(struct type *t);
int			is_array_type(struct type *t);
int			is_integral_type(struct type *t);
int			is_basic_agg_type(struct type *t);
int			is_arr_of_ptr(struct type *t);
int			is_nullptr_const(struct token *, struct type *);
int			type_without_sign(int code);
struct token		*const_from_type(struct type *ty, int from_alignment);
struct token		*const_from_value(void *value, struct type *ty);
int	check_types_assign(
	struct token *t,
	struct type *left,
	struct vreg *right,
	int to_const_ok);
struct type		*addrofify_type(struct type *t);

#endif

