/*
 * Copyright (c) 2004 - 2007, Nils R. Weller
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * Expression parser
 */
#include "expr.h"
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include "token.h"
#include "error.h"
#include "misc.h"
#include "defs.h"
#include "decl.h"
#include "scope.h"
#include "type.h"
#include "icode.h"
#include "debug.h"
#include "cc1_main.h"
#include "subexpr.h"
#include "functions.h"
#include "analyze.h"
#include "symlist.h"
#include "backend.h"
#include "typemap.h"
#include "libnwcc.h"
#include "reg.h"
#include "n_libc.h"

static int
value_is_nonzero(struct tyval *cv) {
	int		size;
	int		i;
	unsigned char	*uc;

	size = cross_get_sizeof_type(cv->type);
	for (uc = (unsigned char *)cv->value, i = 0; i < size; ++i, ++uc) {
		if (*uc != 0) {
			return 1;
		}
	}
	return 0;
}

/*
 * XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
 * doesn't handle cross-compilation :(
 * address->diff should probably become a ``struct tyval'' and then we
 * need a nice set of utility functions to convert host values from/to
 * target-independent values. tyval is probably our best bet for
 * abstracting these things because exec_op() and such already use it
 * in the interface
 */
static void					
addr_add_sub(struct tyval *tv, struct expr *ex, struct type *ty, int op) {
	long		val;
	int		esize = backend->get_sizeof_elem_type(ty);

	cross_do_conv(ex->const_value, TY_LONG);
	/* XXX what to do here? :( */
	ex->const_value->type->code = TY_LONG;
#if 0
	val = *(long *)ex->const_value->value * esize;
#endif
	val = (long)cross_to_host_size_t(ex->const_value);
	val *= esize;

	if (op == TOK_OP_PLUS) {
		tv->address->diff += val;
	} else {
		tv->address->diff -= val;
	}	
}

static struct tyval
do_eval_const_expr(struct expr *tree) {
	struct tyval		ret;
	struct tyval		tmp1;
	struct tyval		tmp2;
	struct tyval		*tv;
	struct tyval		*left;
	struct tyval		*right;
	size_t			i;
	int			rc = 0;
	int			nullok;
	static struct tyval	nulltv;

	ret = nulltv;
	tmp1 = nulltv;
	tmp2 = nulltv; 
	
	if (tree->op == 0) {
		if (tree->data->is_expr) {	
			ret = do_eval_const_expr(tree->data->is_expr);
		} else if (tree->data->meat != NULL) {
			struct token	*t = tree->data->meat;
			
			if (t->type == TOK_STRING_LITERAL) {
				ret.str = t->data;
				ret.type = ret.str->ty;
			} else if (t->type == TOK_IDENTIFIER) {		
				struct decl	*d;

				if ((d = access_symbol(curscope,
					t->data, 1)) == NULL) {
					errorfl(t,"Undeclared identifier `%s'",
						t->ascii);
					ret.type = ret.value = NULL;
					return ret;
				}	
				if (d->dtype->storage != TOK_KEY_STATIC
					&& d->dtype->storage != TOK_KEY_EXTERN
					&& d->dtype->tenum == NULL
					&& (d->dtype->tlist == NULL
					|| d->dtype->tlist->type
						!= TN_FUNCTION)) {
					/*
					 * XXX seems totally bogus.. only
					 * functions should be allowed at for
					 * *values*???
					 */
					errorfl(t, "Use of non-static "
						"identifier `%s' in constant"
						" expression", d->dtype->name);
					ret.type = ret.value = NULL;
					return ret;
				}
				ret.type = d->dtype;
				if (d->dtype->tlist == NULL
					&& d->dtype->tenum != NULL) {
					ret.value = d->vreg->from_const->data;
				} else {
					int	is_addr = 0;

					if (tree->data->operators[0] != NULL
						&& tree->data->operators[0]
						->type
						== TOK_OPERATOR) {
						if (*(int *)tree->data->
							operators[0]->data
							== TOK_OP_ADDR) {
							is_addr = 1;
						}	
					}

					if (!is_addr
						&& (d->dtype->tlist == NULL
						|| d->dtype->tlist->type
						== TN_POINTER_TO)) {
						/*
						 * Just copy initializer for
						 * static variable
						 */
						ret.static_init = d->init;
						if (d->init && d->init->type
							== INIT_EXPR
&& ((ret.type->code != TY_STRUCT && ret.type->code != TY_UNION)
	|| ret.type->tlist != NULL)) {
struct expr *ex = d->init->data;
ret.value = ex->const_value->value;
						}
					} else {
						/* Address constant */
						if (is_addr) {
							ret.type =
								addrofify_type
								(ret.type);
						}

						ret.address =
							n_xmalloc(sizeof *
							ret.address);
						ret.address->dec = d;
						ret.address->diff = 0;
					}
				}
			} else {
				/* Must be constant */
				ret.type = make_basic_type(t->type);
				ret.type = n_xmemdup(ret.type,
					sizeof *ret.type);

				if (IS_FLOATING(ret.type->code)) {
					struct ty_float	*fc;

					fc = t->data;
					ret.value = fc->num->value;
				} else {
					ret.value = t->data;
				}	
				ret.is_nullptr_const =
					is_nullptr_const(t, ret.type);
			}
			ret.alloc = 0;
		} else if (tree->data->is_sizeof) {
			unsigned long	size;
			struct vreg	*vr;

			vr = tree->data->is_sizeof->data;
			ret.type = vr->type;
			size = vr->size;
			ret.value = vr->from_const->data;
		} else {
			abort();
		}

		/* Now apply all unary/postfix/prefix operators */
		tv = &ret;
		for (i = 0; tree->data->operators[i] != NULL; ++i) {
			struct token	*op;
			struct decl	*d;
			int		opval;

			op = tree->data->operators[i];
			if (ret.value == NULL
				&& ret.address == NULL
				&& op->type != TOK_OP_CAST) {
				errorfl(op,
					"`%s' operator used with incorrect "
					"type", op->ascii);
				ret.type = NULL;
				return ret;
			}

			switch (op->type) {
			case TOK_OP_CAST:	
				d = op->data;
				if (tv->is_nullptr_const) {
					if (d->dtype->tlist != NULL
						&& d->dtype->tlist->type ==
						TN_POINTER_TO
						&& d->dtype->tlist->next
						== NULL
						&& d->dtype->code == TY_VOID) {
						/* Remains nullptr */
						;
					} else {
						tv->is_nullptr_const = 0;
					}
				}	
				if (d->dtype->tlist != NULL) {
					if (tree->data->meat &&
						tree->data->meat->type
						== TOK_STRING_LITERAL) {
						;
					} else {
#if 0
						do_conv(tv, TY_ULONG); /* XXX */
#endif
						if (ret.type->tlist == NULL) {
						cross_do_conv(tv, TY_ULONG); /* XXX */
						tv->type = make_basic_type(
							TY_ULONG);
						tv->type = n_xmemdup(tv->type,
							sizeof *tv->type);
						}	
					}
				} else {
					cross_do_conv(tv, d->dtype->code);
				}	
				ret.type = d->dtype;
				break;
			case TOK_ARRAY_OPEN:
				if (ret.address != NULL
					&& (op=tree->data->operators[i+1])
						!= NULL
					&& op->type == TOK_OPERATOR
					&& *(int *)op->data == TOK_OP_ADDR) {
					/* make &foo[x] work */
					struct expr	*ex
						= tree->data->operators[i++]->
						data;

					if (eval_const_expr(ex) != 0) {
						ret.type = NULL;
						return ret; 
					}
					addr_add_sub(&ret, ex, tv->type,
						TOK_OP_PLUS);

#if 0
					do_conv(ex->const_value, TY_ULONG);
					ret.address->diff
						+= *(unsigned long *)ex->
							const_value->value *
						backend->get_sizeof_elem_type
						(tv->type);
#endif
#if 0
					tv->type->tlist->type
						= TN_POINTER_TO; /* XXX */
#endif
				} else {	
					errorfl(tree->data->operators[i],
						"Subscript operator used in"
						" constant expression");
					ret.type = NULL;
					return ret;
				}
				break;
			case TOK_OP_STRUMEMB: /* XXX !??why not TOK_OPERATOR */
			case TOK_OP_STRUPMEMB:
				if (ret.type == NULL
					|| ((ret.type->code != TY_STRUCT
					&& ret.type->code != TY_UNION)
					|| ret.type->tlist != NULL)) {
					errorfl(tree->data->operators[i],
						"Incorrect type for "
						"operator `%s'",
						tree->data->operators[i]->ascii);
					ret.type = NULL;
					return ret;
				}	
				if (ret.address != NULL
					&& (op=tree->data->operators[i+1])
						!= NULL
					&& op->type == TOK_OPERATOR
					&& *(int *)op->data == TOK_OP_ADDR) {
					/* Make &foo.bar / &foo->bar work */
				} else {
					errorfl(tree->data->operators[i],
						"`%s' operator used in"
						" constant expression",
						tree->data->operators[i]->ascii);
					ret.type = NULL;
					return ret;
				}	
				break;
			case TOK_OPERATOR:
				opval = *(int *)op->data;
				if (opval == TOK_OP_ADDR) {
					if (i == 0) {
						; /* already taken care of */
					} else {
						errorfl(op,
							"Parse error at `%s'",
							op->ascii);
						ret.type = NULL;
						return ret;
					}	
#if 0
				} else if (opval == TOK_OP_STRUMEMB) {
					unimpl();
				} else if (opval == TOK_OP_STRUPMEMB) {
					unimpl();
#endif
				} else if (opval == TOK_OP_LNEG
					|| opval == TOK_OP_BNEG
					|| opval == TOK_OP_UPLUS
					|| opval == TOK_OP_UMINUS) { 
					int	oldtype = op->type; /* XXX needed? */
					static struct tyval	tytmp;

					op->type = opval;
					cross_exec_op(op, &tytmp, tv, NULL);
					op->type = oldtype;
					*tv = tytmp;
				} else {
					printf("unknown operator %d\n",
						opval);
					abort();
				}	
				break;
			case TOK_KEY_SIZEOF:
			case TOK_KEY_ALIGNOF:	
				; /* taken care of already */
				break;
			default:
				printf("unknown operator %d\n", op->type);
				unimpl();
			}
		}
		return ret;
	}

	tmp1 = do_eval_const_expr(tree->left);

	if (tree->op == TOK_OP_LAND) {
		int	res = 0;

		if (value_is_nonzero(&tmp1)) {
			tmp2 = do_eval_const_expr(tree->right);
			if (value_is_nonzero(&tmp2)) {
				res = 1;
			}
		}
		ret.type = make_basic_type(TY_INT);
		ret.type = n_xmemdup(ret.type, sizeof *ret.type);
		ret.value = n_xmemdup(&res, sizeof res);
		goto out;
	} else if (tree->op == TOK_OP_LOR) {	
		int	res = 0;

		if (!value_is_nonzero(&tmp1)) {
			tmp2 = do_eval_const_expr(tree->right);
			if (value_is_nonzero(&tmp2)) {
				res = 1;
			}
		} else {
			res = 1;
		}
		ret.type = make_basic_type(TY_INT);
		ret.type = n_xmemdup(ret.type, sizeof *ret.type);
		ret.value = n_xmemdup(&res, sizeof res);
		goto out;
	} else if (tree->op != TOK_OP_COND) {
		/* XXX broken */
		tmp2 = do_eval_const_expr(tree->right);
		nullok = 0;
	} else {
		/* Is TOK_OP_COND */
		if (tree->right->op != TOK_OP_COND2) {
			errorfl(tree->right->tok,
				"Parse error - expected second part "
				"of conditional operator");
			ret.type = NULL;
			return ret;
		}
		if (value_is_nonzero(&tmp1)) {
			tmp2 = do_eval_const_expr(tree->right->left);
		} else {
			tmp2 = do_eval_const_expr(tree->right->right);
		}
		if (tmp2.type == NULL) {
			return ret;
		}
		if (tmp2.type->code < TY_INT) {
			/*
			 * XXX we also need some typechecking ... And
			 * usual arithmetic conversions!!! E.g.
			 * foo < bar? 0ll: 0;
			 * ... should convert 0 to 0ll
			 */
			cross_do_conv(&tmp2, TY_INT);
		}
		ret.type = tmp2.type;
		ret.value = tmp2.value;
		goto out;
	}

	tv = &ret;
	left = &tmp1;
	right = &tmp2;

	if (tmp1.type == NULL
		|| (!nullok && tmp2.type == NULL)
		|| (!nullok && (rc = cross_convert_tyval(&tmp1, &tmp2)) != 0)) {
		ret.type = NULL;
		ret.value = NULL;
		if (rc != 0) {
			/*
			 * XXX mm this is kludged, convert_tyval() shouldn't
			 * return an error in the first place ;/
			 */
			if (tmp1.address || tmp2.address) {
				if (tree->op == TOK_OP_PLUS) {
					static struct expr	dum;

					/* Must be addr + integer */
					if (tmp1.address == NULL) {
						dum.const_value = &tmp1;
						addr_add_sub(&tmp2, &dum,
							tmp2.type,
							TOK_OP_PLUS);
						return tmp2;
					} else if (tmp2.address == NULL) {
						dum.const_value = &tmp2;
						addr_add_sub(&tmp1, &dum,
							tmp1.type,
							TOK_OP_PLUS);
						return tmp1;
					}
				} else if (tree->op == TOK_OP_MINUS) {
					unimpl();
				}
			}
			errorfl(tree->tok, "Incompatible types in expression");
			ret.type = NULL;
			return ret;
		}
		return ret;
	}
	
	if ((left->address == NULL && left->value == NULL)
		|| (right->address == NULL && right->value == NULL)) {
		if (tree->op != TOK_OP_COND) {
			errorfl(tree->right->tok, "Incompatible types for "
				"operator");	
			ret.type = NULL;
			return ret;
		}	
	}	

	switch (tree->op) {
	case TOK_OP_PLUS:
		if (left->address || right->address) {
			unimpl();
		} else {	
			cross_exec_op(tree->tok, tv, left, right); 
		}	
		break;
	case TOK_OP_MINUS:
		if (left->address || right->address) {
			unimpl();
		} else {	
			cross_exec_op(tree->tok, tv, left, right); 
		}	
		break;
	case TOK_OP_MULTI:
		cross_exec_op(tree->tok, tv, left, right); 
		break;
	case TOK_OP_DIVIDE:
		cross_exec_op(tree->tok, tv, left, right); 
		break;
	case TOK_OP_SMALL:
		cross_exec_op(tree->tok, tv, left, right); 
		break;
	case TOK_OP_GREAT:
		cross_exec_op(tree->tok, tv, left, right); 
		break;
	case TOK_OP_GREATEQ:
		cross_exec_op(tree->tok, tv, left, right); 
		break;
	case TOK_OP_SMALLEQ:
		cross_exec_op(tree->tok, tv, left, right); 
		break;
	case TOK_OP_LEQU:
		cross_exec_op(tree->tok, tv, left, right); 
		break;
	case TOK_OP_LNEQU:
		cross_exec_op(tree->tok, tv, left, right); 
		break;
	case TOK_OP_MOD:
		cross_exec_op(tree->tok, tv, left, right); 
		break;
	case TOK_OP_BAND:
		cross_exec_op(tree->tok, tv, left, right); 
		break;
	case TOK_OP_BOR:
		cross_exec_op(tree->tok, tv, left, right); 
		break;
	case TOK_OP_BXOR:
		cross_exec_op(tree->tok, tv, left, right); 
		break;
	case TOK_OP_BSHL:
		cross_exec_op(tree->tok, tv, left, right); 
		break;
	case TOK_OP_BSHR:
		cross_exec_op(tree->tok, tv, left, right); 
		break;
	default:
		printf("Badness to the maximum %d\n", tree->op);
		abort();
	}

	if (tree->op != TOK_OP_COND) {
		free(left->value);
		free(right->value);
	}	

out:
	return ret;
}

int
eval_const_expr(struct expr *ex) {
	static struct tyval	nulltv;
	struct tyval	tv;
	struct tyval	*dtv;

	if (ex->const_value != NULL) {
		/* already been evaluated */
		return 0;
	}	
	tv = nulltv;

	tv = do_eval_const_expr(ex);
	if (tv.type == NULL) {
		return -1;
	} else {
		rv_setrc_print(tv.value, tv.type->code, 0);
		dtv = n_xmalloc(sizeof tv);
		memcpy(dtv, &tv, sizeof tv);
		ex->const_value = dtv;
	}

	return 0;
}

